#!/usr/bin/env/python
# -*- coding: utf-8 -*-

import os
import time
import logging
import subprocess
import traceback

from rospeex_core.exceptions import WebAudioMonitorException

logger = logging.getLogger(__name__)

class WebAudioMonitor(object):
    """ AudioMonitor launcher class """

    def __init__(self, app_dir, app_name, send_port, recv_port, websocket_port):
        """
        initialize AudioMonitor class
        @param app_dir:
        @type  app_dir:
        @param app_name:
        @type  app_name:
        @param send_port:
        @type  send_port: int
        @param recv_port:
        @type  recv_port: int
        @param websocket_port:
        @type  websocket_port: int
        @raises WebAudioMonitorException
        """
        # set init setting
        self._process = None
        self._send_port = send_port
        self._recv_port = recv_port
        self._websocket_port = websocket_port
        self._app_dir = app_dir

        # get AudioMonitor path
        try:
            self._app_path = os.path.join(app_dir, app_name)
        except Exception, e:
            msg = 'monitorServer is not found. msg: %s'%str(e)
            raise WebAudioMonitorException(msg)

        # check rospack result
        if not os.path.exists(self._app_path):
            msg = 'rospeex is not found. msg: %s'%str(self._app_path)
            logger.warn(msg)
            raise WebAudioMonitorException(msg)

        # show args
        logger.info(
            'create audio monitor : \n\
            \t send_port        :%d\n\
            \t recv_port        :%d\n\
            \t websocket_port   :%d\n\
            \t application_path :%s'
            %(send_port, recv_port, websocket_port, self._app_path))


    def launch(self):
        """
        launch AudioMonitor
        @raises WebAudioMonitorException
        """
        if self._process is None:
            # launch process
            args = [self._app_path,
                    str(self._websocket_port),
                    str(self._recv_port),
                    str(self._send_port)]
            options = {'cwd': self._app_dir}

            try:
                # launch subprocess
                self._process = subprocess.Popen(args, **options)
            except OSError as err:
                self._process = None
                raise WebAudioMonitorException('Failed to launch. msg:', str(err))
            except Exception as err:
                self._process = None
                msg = 'Failed to launch. Unknown exception. msg: %s Traceback: %s' % (str(err), traceback.format_exc())
                raise WebAudioMonitorException(msg)
            else:
                logger.debug('AudioMonitor successfuly launched.')
        else:
            logger.info('AudioMonitor is aleady launched. pid:%d'%self._process.pid)


    def exit(self):
        """ exist AudioMonitor """
        if self.is_alive():
            # kill process
            self._process.kill()
            self._process.wait()
            self._process = None
        else:
            logger.info("AudioMonitor is aleady exit.")
            self._process = None


    def is_alive(self):
        """
        check audio monitor alive
        @return: True for audio monitor is alive
                 False for audio monitor is NOT alive
        @rtype : Bool
        """
        if self._process is not None:
            if self._process.poll() is None:
                return True
        else:
            logger.info('AudioMonitor is not launched.')

        return False


if __name__ == '__main__':
    send_port = 16001
    recv_port = 5002
    web_socket_port = 9000

    try:
        audio_monitor = WebAudioMonitor(send_port, recv_port, web_socket_port)
        audio_monitor.launch()
    except WebAudioMonitorException as err:
        print str(err)
    except Exception as err:
        print str(err)

    while audio_monitor.is_alive():
        time.sleep(0.1)

    audio_monitor.exit()

